'' ILI9325 driver using the Touchblade161 design for faster ram to display transfers
'' Average Joe and James Moxham 2012

CON


  _clkmode      = xtal1 + pll16x                        ' use crystal x 16
  _xinfreq      = 5_000_000

  _1ms    = 1_000_000 /     1_000                         ' Divisor for 1 ms   


' SD pins
  _SD_DO = 21
  _SD_CLK = 22
  _SD_DI = 23
  _SD_CS = 24                                          
  _SD_WP = -1                                           ' -1 ifnot installed.
  _SD_CD = -1                                           ' -1 ifnot installed.
  _RTC_DAT = -1                                         ' -1 ifnot installed. For kye's sd driver
  _RTC_CLK = -1                                         ' -1 ifnot installed.

' touchscreen pins
  Touch_Clock = 0                        
  'Touch_ChipSelect = done with a 138 select
  Touch_DataIn = 1            
  Touch_DataOut = 2
            
' general purpose constants
  'Margin = 4                                            ' some characters have xoffset of -1 or -2 eg v, and won't display on the extreme left edge

  
' external ram size of data blocks
  Ram_PicSize = 76800                                   ' background image desktop (words)
  Ram_Fontsize = 65536                                  ' 128k for fonts (largest is around 108k)
  Ram_MaskSize = (64*64*2)/2                            ' 64x64 (59x60) bytes divide by 2 for words
  Ram_IconSize = (64*64*2)/2                            ' 2 bytes per pixel converts from .raw in the routine
  Ram_TextSize = 32768                                  ' 64k for text
  Ram_Pic2Size = 76800                                  ' picture for program called from desktop
  Ram_Pic8kSize = 8192                                  ' small general purpuse picture buffer
  Ram_Pic1kSize = 1024                                  ' tiny general purpose picture buffer

' external ram locations, if adding more, follow the pattern
  Desktop = 0                                           ' location of picture                                    
  FontTable = Desktop+Ram_PicSize                       ' address for ram chip (2 bytes per address)
  Mask = FontTable + Ram_Fontsize                       ' address for raw mask image
  Icon = Mask + Ram_MaskSize                            ' address for icon image
  Text = Icon + Ram_IconSize                            ' text buffer for word processing etc
  Picture = Text + Ram_TextSize                         ' picture for program called from desktop (preserve desktop so quicker redraw)
  Pic1 = Picture + Ram_Pic2Size                         ' small 8kiloword picture buffer
  Pic2 = Pic1 + Ram_Pic8kSize                           ' small 8kiloword picture buffer
  Pic3 = Pic2 + Ram_Pic8kSize                           ' tiny 1kiloword picture buffer
  Pic4 = Pic3 + Ram_Pic1kSize                           ' tiny 1kiloword picture buffer  
  Pic5 = Pic4 + Ram_Pic1kSize                           ' tiny 1kiloword picture buffer  
  Pic6 = Pic5 + Ram_Pic1kSize                           ' tiny 1kiloword picture buffer  

OBJ

  fat:           "SD3.01_FATEngine.spin"                 ' thanks to Kye
  spi:           "SPI_ASM"                               ' thanks to Beau
  str:           "ASCII0_STREngine.spin"                 ' thanks to Kye   

VAR
   long orientation
   long curx, cury
   long clkcycles               ' for the delay routine
   word ScreenWidth             ' either 240 in portrait or 320 in landscape
   word ScreenHeight            ' 320 in portrait or 240 in landscape
   byte LineOfText1[20]         ' general purpose string buffer
   byte LineOfText2[20]         ' general purpose string buffer
   byte FontHeight              ' size of the current loaded font (pixels = fontsize *.75)
   byte Margin
   word BackFontColor           ' the background color in RRRRRGGG GGGBBBBB format
   
                  
   'byte sdbuffer[512]     ' can't put these here, writes to sdbuffer sometimes back overwrite variables above this eg fontheight!!
   'byte buffer2[512]      ' so put these in a dat section
   'word rambuffer[256]
   
   
PUB Main |touchtest  ' debug value
  Margin := 4                                           ' if set to zero some letters don't display on the left side eg v
  SelectMemGroup                                        ' select the memory group
  SetOrientation(true)                                  '  in portrait (true) or landscape mode (false) 
  start_ram                                             ' start the external ram driver in a cog
  Start_ILI9325                                         ' start the display
  Clearscreen(RGBtoWord(0,0,0))                         ' clear screen to black
  Propfont_string(string("SD"))                         ' string to send, uses internal prop font so can see something if the SD fails to mount   
  ifnot( fat.FATEngineStart(_SD_DO, _SD_CLK, _SD_DI, _SD_CS, _SD_WP, _SD_CD, _RTC_DAT, _RTC_CLK, -1))   
    reboot
  fat.mountPartition(0)                                 ' mount the sd card
  'Bootmessage                                           ' print "booting" using a font from the sd card
  
  LoadRawPicture(string("balloon.raw"),Desktop)         ' display this .raw image as the background
  LoadIcons                                             ' load the mask and icons onto the screen

  ILISetcursor(0,0)                                     ' cursor for debugging to 0,0
  SelectSPIGroup                                        ' selects this group and starts the cog 
  Touchscreen                                           ' wait for input

PUB Buttons | ck1x,ck1y,ck1s,ck2x,ck2y,ck2s,rd1x,rd1y,rd1s,rd2x,rd2y,rd2s,rd3x,rd3y,rd3s,bu1x,bu1y,xval,yval,redraw ' test out buttons, radio buttons, checkboxes 
   SetOrientation(true)                                 ' portrait view
   clearscreen(RGBtoWord(212,208,200))                  ' background color determined from paintshop
   ILIloadfont(string("MSans14.ifn"))                   ' 14 point seems to be the same as 8.25 in vb.net ? why
   LoadSmallPicture(string("butn8430.raw"),84,30,Pic1)  ' a button 84x30 in size
   LoadSmallPicture(string("Check13A.raw"),13,13,Pic3)  ' unchecked checkbox pic3
   LoadSmallPicture(string("Check13B.raw"),13,13,Pic4)  ' checked checkbox pic4
   LoadSmallPicture(string("Radio12A.raw"),12,12,Pic5)  ' unchecked radiobox pic 5
   LoadSmallPicture(string("Radio12B.raw"),12,12,Pic6)  ' checked radiobox pic 5
   redraw := true    
   ' set all the locations of the objects on the screen
   ck1x := ck2x := rd1x := rd2x := rd3x := 20           ' x all the same 
   ck1y := 10    
   ck2y := ck1y + 40                                    ' y coordinates
   rd1y := ck2y + 40
   rd2y := rd1y + 40
   rd3y := rd2y + 40
   bu1x := 80                                           ' button position
   bu1y := 280
   ck1s := true                                         ' startup status
   ck2s := false
   rd1s := true
   rd2s := false
   rd3s := false
   ' now wait for touch
   SelectSPIGroup                                       ' talk to the spi touchscreen
   repeat
      if redraw                                         ' if redraw true then redraw all the objects
         SelectMemGroup                                 ' display mode
         DrawCheckbox(ck1x,ck1y,ck1s,string("Checkbox1"))         ' checkbox
         DrawCheckbox(ck2x,ck2y,ck2s,string("Checkbox2"))         ' checkbox
         DrawRadiobox(rd1x,rd1y,rd1s,string("Radiobox1"))         ' radiobox
         DrawRadiobox(rd2x,rd2y,rd2s,string("Radiobox1"))         ' radiobox 
         DrawRadiobox(rd3x,rd3y,rd3s,string("Radiobox2"))         ' radiobox
         DrawButton(bu1x,bu1y,pic1,string("Ok"))                  ' button (fixed size at the moment)
         redraw := false
         SelectSPIGroup                                 ' talk to the spi touchscreen  
      yval := TouchYPercent                             ' decode yval 0-100%                
      xval := TouchXPercent                             ' decode xval 0-100%
      if (xval <> 255)  and (yval <> 255)               ' valid keypress
        xval := (xval * 24) / 10                        ' scale to pixels as can measure from the picture in paintshop
        yval := (yval * 32) / 10                        ' scale to pixels 
        ' cheat a bit here, no need to look for x value as only one column of objects. See calculator for x and y buttons
        case yval
          (ck1y-10)..(ck1y+25): !ck1s                   ' checkbox is small so active area bigger than the checkbox
                                redraw := true          ' redraw the screen
          (ck2y-10)..(ck2y+25): !ck2s                   ' checkbox is small so active area bigger than the checkbox
                                redraw := true         
          (rd1y-10)..(rd1y+25): rd1s :=   true            ' there probably is a smarter way to make one value true and all the rest false
                                rd2s :=   false
                                rd3s :=   false                   
                                redraw := true
          (rd2y-10)..(rd2y+25): rd1s :=   false            
                                rd2s :=   true
                                rd3s :=   false                   
                                redraw := true
          (rd3y-10)..(rd3y+25): rd1s :=   false            
                                rd2s :=   false
                                rd3s :=   true                  
                                redraw := true     
          bu1y..(bu1y + 30): return                     ' button so exit this screen 
        pause1ms(250)                                   ' debounce
   repeat

PUB DrawCheckbox(x,y,Checked,Stringptr) | i                ' draw a checkbox
   if checked
     i := pic4
   else
     i := pic3
   DrawSmallPicture(x,y,13,13,i)               ' size determined from original picture
   curx := x+20                                         ' add 20 width
   cury := y
   ILIPrintString(stringptr)

PUB DrawRadioBox(x,y,Checked,Stringptr)| i              ' draw a radio box
   if checked
     i := pic6
   else
     i := pic5
   DrawSmallPicture(x,y,12,12,i)               ' size determined from original picture
   curx := x+20                                         ' add 20 width
   cury := y
   ILIPrintString(stringptr)

PUB DrawButton(x,y,Ramaddress,Stringptr)                ' draw a radio box
   DrawSmallPicture(x,y,84,30,Ramaddress)               ' size determined from original picture (later build variable sized buttons, or load them from the sd card)
   curx := x + (84/2)-(str.len(stringptr)*4 / 2)        ' centre, avg width 4 per char 84 could be a variable width passed to this routine
   cury := y + 8
   ILIPrintString(stringptr)         

PUB Keyboard | capslock, backspace,xval,yval,address,ascii
   Clearscreen(%00000000_00000000)                      ' clear the screen
   SetOrientation(false)                                ' landscape view
   ILIloadfont(string("Arial30.ifn"))                   ' load this font
   RawSDToDisplay(string("KeyL.raw"))                   ' load this picture            
   capslock := 0 ' default to off                       ' caps lock 
   backspace := 0 'default to off                       ' backspace is zero
   str.builtstring(true)                                ' clear the working string
   SelectSPIGroup                                       ' talk to the spi touchscreen
   ILISetcursor(margin,27)                              ' cursor to top left
   repeat
      xval := TouchYPercent                           ' decode yval 0-100%  and swap x and y as this is landscape mode              
      yval := TouchXPercent                          ' decode xval 0-100%
      if (xval <> 255)  and (yval <> 255)               ' valid keypress
        SelectMemGroup                                  ' displaying now
        address :=255
        ascii := 255
        xval := (xval * 32) / 10                        ' scale to pixels as can measure from the picture in paintshop
        yval := 240 - ((yval * 24) / 10)                ' scale to pixels and org top left                        
        case yval
          133..162:address := xval / 32                 ' top row qwertyuiop
          163..188:address := 10 + ((xval-17) / 32)     ' second row addres 10 up, offset
          189..214:address := 19 + ((xval -47) / 32)    ' third row, includes two buttons in one
                   if xval<47
                     address := 255                     ' non printing character capslock
                     !capslock                          ' toggle capslock
                   if xval >273
                     !backspace                         ' backspace character
          215..239:if xval >64 and xval < 256
                      address := 26                     ' space bar 
                   if xval>256
                      address :=27                      ' carriage return (LF added below)   
        if address < 255 and backspace == 0
           ascii := lookupz(address: "Q","W","E","R","T","Y","U","I","O","P","A","S","D","F","G","H","J","K","L","Z","X","C","V","B","N","M",32,13)
          if (capslock == 0) and (ascii => 65) and (ascii =< 90) 
            ascii += 32                                 ' lower case              
          curx := ILIChar(ascii,curx,cury)              ' only print if a valid ascii character
          str.buildstring(ascii)
          if ascii == 13                                ' add line feed if this is a return
            ascii := 10                                 ' add LF
            curx := ILIChar(ascii,curx,cury)            ' display line feed
            str.buildstring(ascii)                      ' and add to the built string
        if backspace <>0                                ' backspace redraws the entire string buffer
          backspace := 0                                ' backspace now zero
          str.buildstring(8)                            ' kye's string routine removes a character
          ClearRectangle(0,22,319,132,BackFontColor)  ' clear an area the same color as the font background        
          ILISetcursor(margin,27)                       ' top left of the screen
          ILIPrintstring(str.builtstring(false))        ' redraw text
        
        if xval => 0 and xval <45 and yval => 0 and yval <21
          return                                        ' cancel button
        if xval => 280 and xval < 320 and yval => 0 and yval <21
          result := str.builtstring(true)                   ' the 'send' button returns the string including carriage return, line feeds (can do multiple lines of input)
          return
        pause1ms(250)                                   ' debounce
        SelectSPIGroup                                  ' wait for next keypress



PUB PhotoAlbum | picturenumber
   picturenumber := 1
   repeat until picturenumber > 5
     if picturenumber == 1
        RawSDToDisplay(string("cat.raw"))         ' draw a portrait picture
     if picturenumber ==2
        SetOrientation(false)                           ' draw a landscape picture
        RawSDToDisplay(string("wallaby.raw"))
        SetOrientation(true)     
     if picturenumber == 3
        RawSDToDisplay(string("kluger.raw"))
     if picturenumber == 4
        RawSDToDisplay(string("elle.raw"))    
     if picturenumber == 5
        RawSDToDisplay(string("balln.raw"))
     result := Touchswish
     if result > 0
       picturenumber --
     else
       picturenumber ++
   clearscreen(%00000000_00000000)    
   pause1ms(1000) ' otherwise still touching the screen    
PUB RawSDToDisplay(stringptr)                 ' send straight to the display
    Clearscreen(%00000000_00000000)
    LoadRawPicture(stringptr,picture) 
    DrawPicture

PUB TouchSwish | xval,oldxval,difference,average ' returns result = +ve or -ve depending on swish direction     ' did finger move left or right?
      SelectSPIGroup ' turn on the spi touchscreen
      oldxval := 255' for startup
      average := 0
      repeat until (average > 30) or (average < -30)
        ' possibly add a decay function here for average
        pause1ms(50)
        xval := TouchXPercent                        ' decode xval 0-100%
        if xval == 255                                  ' not touched
          oldxval := 255                                ' reset oldxval as well
        else
          if oldxval <>255                                ' discard the first reading as the screen is touched
            difference := xval - oldxval
            'printdecimal(xval)
            'printdecimal(difference)
            average := average + difference             ' 
            'printdecimal(average)
            'crlf
          oldxval := xval                                 ' store for next time
      result := average                                 ' return negative or positive value         

PUB TouchStart
      SPI.start(3,0)                                        ' delay,state, start clock high

PUB TouchX
      SPI.SHIFTOUT(Touch_DataIn, Touch_Clock, 5, 8 , %1101_0000)  ' reads x from  500 to 3700  (off < 500 )
      result := SPI.SHIFTIN(Touch_DataOut,Touch_Clock,2, 12)

PUB TouchY      
      SPI.SHIFTOUT(Touch_DataIn, Touch_Clock, 5, 8 , %1001_0000)  ' reads y from 400 to 3800 (off > 3800 ) 
      result := SPI.SHIFTIN(Touch_DataOut,Touch_Clock,2, 12)    

PUB TouchValue | xval,yval,x,y ' returns % values 0-100% (or 255 for no touch)
      xval := TouchX
      yval := TouchY
      if (xval => 500) and (yval =< 3800)                       ' both have to be valid for a touch
        pause1ms(2)                               ' tiny delay then read again as the first reads may not have the correct value
        xval := TouchX
        yval := TouchY
        if (xval => 500) and (yval =< 3800)                 ' and if valid again, now more likely this value is correct
            x := xval - 440                                    ' scale 0-3360
            x /=  32                                     ' scale 0-100
            x <#= 100                                    ' max 100
            x #>=0                                       ' min 0
            x := 100 - x                                 ' so left edge on left
            y := yval -420                               ' scale 0-3180
            y /= 32                                      ' scale 0-100
            y <#= 100                                    ' max 100
            y #>=0                                       ' min 0
            y := 100-y                                   ' so 0 is top left     
            result := (y <<8) | x                       ' return 00000000_00000000_YYYYYYYY_XXXXXXXX   
      if (xval < 500) or (xval >3800)
        result := $0000FFFF                             ' invalid number(s)

PUB TouchXPercent
    result := TouchValue & 255                          ' decode xval 0-100%

PUB TouchYPercent
    result := TouchValue >> 8                           ' decode yval 0-100%           

PUB LoadIcons                                           ' load all the icons onto the screen. Possibly use a .ini text file for this
  ReadMask                                              ' read the mask.raw file into external memory 
  MergeBackgroundIcon(string("photos.raw"),0,120)
  MergeBackgroundIcon(string("calc.raw"),60,120)        ' draw an icon onto the background
  MergeBackgroundIcon(string("facebook.raw"),120,120)   ' draw an icon
  MergeBackgroundIcon(string("videos.raw"),180,120)     ' draw an icon
  MergeBackgroundIcon(string("notes.raw"),0,180)        ' draw an icon
  MergeBackgroundIcon(string("80x40.raw"),60,180)       ' skype icon for 80x40 tiny font demo
  MergeBackgroundIcon(string("music.raw"),120,180)      ' music icon for testing sliders
  MergeBackgroundIcon(string("prop.raw"),180,180)       ' prop icon for a GUI buttons, radio, checkboxes etc

PUB TouchScreen  | yval, xval,pointer
    repeat
      yval := TouchYPercent                               ' decode yval 0-100%               
      xval := TouchXPercent                              ' decode xval 0-100%
      if (xval <> 255)  and (yval <> 255)
        'hex(xval,2)      ' if display any debug messages need to restart the SPI cog as it has been stopped by the debug display
        'hex(yval,2)
        'repeat
        pointer := (yval/20)*4+(xval/25)       ' so can use a case statement 0-19
        case pointer                                    ' 0 to 19
           0:                   
           1:                   
           2:
           3:
           4:
           5: 
           6:
           7:
           8:  PhotoAlbum
           9:  Calculator                 
           10: LoremPortLand
           11: result := \fat.bootpartition(string("MOVIE.EXE"))     ' reboot and run this program
           12: Keyboard
           13: TinyFont
           14: Slider
           15: Buttons
           16: 
           19: 

        SetOrientation(true)
        SelectMemGroup                                  ' ready for mem out
        DrawDesktop                                     ' redraw the background
        SelectSPIGroup                                  ' restart the spi driver

PUB Slider | Percent,startx,starty,xval,yval
    Clearscreen(%00000000_00000000)                     ' background
    LoadSmallPicture(string("SlideBck.raw"),36,92,Pic1) ' size determined by paint shop pro, original bmp file (.raw loses the size data)
    LoadSmallPicture(string("SlideKnb.raw"),27,19,Pic2) ' slider knob
    startx:=100                                         ' location of the slider        
    starty:=100
    percent := 50                                       ' start percentage
    DrawSmallPicture(startx,starty,36,92,Pic1)          ' size determined from original picture
    DrawSmallPicture(startx+5,starty+3+(percent*67/100),27,19,Pic2)  ' scale for this particular slider image
    repeat until (yval >80 and yval<100)                ' exit condition hit anywhere in the bottom of the screen
      SelectSPIGroup
      yval := TouchYPercent                           ' decode yval 0-100%               
      xval := TouchXPercent                          ' decode xval 0-100%
      if (xval <> 255)  and (yval <> 255)               ' valid keypress 
        SelectMemGroup
        percent := (yval-30)*400/100                    ' offset and scale
        if percent <0                                   ' make sure in range 0-100
          percent :=0
        if percent >100
          percent :=100    
        DrawSmallPicture(startx,starty,36,92,Pic1)      ' size determined from original picture
        DrawSmallPicture(startx+5,starty+3+(percent*67/100),27,19,Pic2)  ' scale for this particular slider image
      pause1ms(10)                                      ' small pause

PUB Calculator | xval,yval,address,ascii,operand
    Clearscreen(%00000000_00000000)                     ' so something happens immediately
    LoadRawPicture(string("calcpic.raw"),Picture)       ' draw the calculator                                
    DrawPicture
    ILIloadfont(string("DSdig.ifn"))
    str.builtstring(true)                               ' clear the working string
    SelectSPIGroup                                      ' talk to the spi touchscreen
    repeat
      yval := TouchYPercent                               ' decode yval 0-100%               
      xval := TouchXPercent                               ' decode xval 0-100%
      if (xval <> 255)  and (yval <> 255)                    ' valid keypress
        address :=0
        case (yval*32)/10                                    ' convert 0-100 to 0-320 as easier to work with pixels off the picture
          60..101: address :=0                               ' row of a 4x6 matrix
          104..140: address :=4
          146..183: address :=8
          187..222: address :=12
          228..265: address :=16
          271..307: address :=20
        case (xval*24)/10                                   ' column convert percent to 0-239
          0..66: address += 0
          67..119: address +=1
          120..172: address +=2
          173..239: address +=3
        ascii := lookupz(address: "X","Q","Q","Q","C",8,"/","*","7","8","9","-","4","5","6","+","1","2","3","=","0","0",".","=")
        case ascii
          "X"             : return                                              ' close the calculator
          "C"             : str.builtstring(true)                               ' clear the current string
                            CalcClear
          "0".."9",8      : str.buildstring(ascii)                              ' build the string. Can also handle backspace
                            CalcClear                            
                            ILIprintstring(str.builtstring(false))              ' reprint the string each new char in case a backspace 
          "+","-","*","/" : str.buildstring(ascii) ' display these too so can see typed something
                            CalcClear
                            ILIprintstring(str.builtstring(false))    
                            str.copy(str.builtstring(true),@lineoftext1)        ' move the string to lineoftext1 and reset
                            operand := ascii
          "=" :             xval := str.decimaltointeger(@lineoftext1)          ' first operand. recycle xval and yval
                            yval := str.decimaltointeger(str.builtstring(true)) ' second operand
                            case operand
                                  "+" : address := xval + yval                  ' add
                                  "-" : address := xval - yval                  ' subtract
                                  "*" : address := xval * yval                  ' multiply
                                  "/" : address := xval / yval                  ' integer division
                            str.str(@lineoftext2,address)                       ' convert answer to string
                            str.trimstring(@lineoftext2) ' if negative, leave the -, if positive remove the leading space
                            CalcClear                                           ' clear the display ready to print the answer
                            ILIprintstring(@lineoftext2)                        ' print the answer
        SelectSPIGroup                                                          ' restart the spi group if display something
        pause1ms(250)' only display once so delay half a second to "debounce"

PUB CalcClear
     ClearRectangle(25,21,220,48,BackFontColor)       ' clear an area the same color as the font background        
     ILISetcursor(25,15)
  

PUB LoremPortLand
    SetOrientation(false)
    Lorem
    SetOrientation(true)
    Lorem


PUB Lorem | filesize,pointer,c,block     
     filesize := Loadfile(string("Lorem2.txt"))                                           ' load in the text to external ram
     ILISetcursor(0,0)                                  ' sets curx to  zero and cury to zero
     ILIloadfont(string("Times20.ifn"))                 'load font
     ClearRectangle(curx,cury,screenwidth,screenheight,BackFontColor)
     curx := Margin                                     ' some letters have an xoffset of -1 and so won't display
     repeat block from 0 to 0                           ' block is 128 words = 256 bytes
       docmd("T",@sdbuffer,Text+(block<<7),256)         ' read text into sdbuffer
       pointer :=0
       repeat 256
          c := byte[@sdbuffer][pointer]
         if c=< 126
           curx := ILIChar(c,curx,cury)                   ' only print if a valid ascii character
         pointer ++
     pause1ms(2000)

PUB TinyFont | filesize,pointer,c,block
     SetOrientation(false)
     Margin := 0    
     filesize := Loadfile(string("ws.txt"))                                           ' load in the text to external ram
     ILISetcursor(0,0)                                  ' sets curx to  zero and cury to zero
     ILIloadfont(string("4x6.ifn"))                 'load font
     ClearRectangle(curx,cury,screenwidth,screenheight,BackFontColor)
     'curx := Margin                                     ' some letters have an xoffset of -1 and so won't display
     repeat block from 0 to 10                           ' block is 128 words = 256 bytes
       docmd("T",@sdbuffer,Text+(block<<7),256)         ' read text into sdbuffer
       pointer :=0
       repeat 256
          c := byte[@sdbuffer][pointer]
         if c=< 126
           curx := ILIChar(c,curx,cury)                   ' only print if a valid ascii character
         pointer ++
    pause1ms(10000)
    Margin := 4                                         ' restore to default value     

     
     
PUB LoadFile(strpointer) | filesize,i                   ' reads a text file into external ram
    fat.openfile(strpointer,"R")                        ' open a file
    filesize := fat.filesize
    i := Text                                           ' store text here
    repeat (filesize >>8 ) + 1                          ' so works with 256 byte blocks
      fat.readdata(@sdbuffer,256)                       ' get 256 bytes from sd card
      docmd("S",@sdbuffer,i,128)                        ' send 128 words to external ram    
      i +=128                                           ' increment counter
    fat.closefile   
    return filesize

PUB NewFont(strpointer)
    ILIloadfont(strpointer)
    ' only call this routine with the cursor curx at zero
    ClearRectangle(curx,cury,screenwidth,cury+FontHeight,BackFontColor)             ' clear an area the same color as the font background  

PUB Bootmessage
    ILIloadfont(string("JDLCD.ifn"))                    ' LED dot matrix type font green on black         
    ILISetcursor(0,0)                                   ' set cursor to this location
    ClearRectangle(0,0,screenwidth,FontHeight,BackFontColor)                 ' clear an area the same color as the font background       
    ILIprintstring(string(">Booting"))  

PUB ILIChar(ascii,x,y) | jump,size,width,height,ramaddress,xoffset,yoffset,xadvance   ' read out a character from a .ifn file stored in ram
  ' pass Fonttable = global
  ' moves to next line if off the end
    jump := ReadRamLong(fonttable,(ascii << 2) + 256)   ' jump location = ascii *4 plus fonttable + 256
    docmd("T",@buffer2,fonttable+jump>>1,20)            ' read in the width height as a block = quicker than readramlong
    size := long[@buffer2][0]                           ' size in pixels of this character
    xadvance := long[@buffer2][5]                       ' amount to move to next character   
    'if size > 0 and ascii <> 32                         ' no need to print anything if size is zero or read more font data
    width := long[@buffer2][1]                        ' width
    height := long[@buffer2][2]                       ' height
    xoffset := long[@buffer2][3]                      ' xoffset to move
    yoffset := long[@buffer2][4]                      ' yoffset to move
    if (x+width -1 + xoffset) > screenwidth
        ILIcrlf                                         ' do a new line if it won't fit
        x := curx
        y := cury
    if ascii < 32
      xadvance := 0                                     ' add nothing if a non displaying character
    if ascii == 13                                  ' carriage return
        ILIcr
        x := curx
    if ascii ==10                                   ' line feed (new line)
        ILIlf
        y := cury
    if ascii > 32 ' space not displayed
      Draw(x+xoffset,y+yoffset,x+width-1+xoffset,y+height-1+yoffset)      ' draw on screen
      RamToDisplay(fonttable+((jump+32)>>1),size)       ' move bytes from ram out to the display
    return x + xadvance

PUB ILIcrlf
    ILIcr
    ILIlf

PUB ILIcr                                               ' carriage return, to left margin
    curx := Margin                                      ' some characters are -1 xoffset so won't display

PUB ILIlf                                               ' line feed, move up one
    cury += FontHeight
    if cury > screenheight
      ClearRectangle(0,0,screenwidth,screenheight,BackFontColor) 
      cury := 0                                         ' if off the bottom of the screen then clear screen start at the top (VT00 does scrolling)
    ' don't clear background here, do in calling routine
    'ClearFontBackground(0,cury,screenwidth,cury + FontHeight) ' clear background line to the background color ready to draw          


PUB ILIPrintstring(stringptr)                 'print at curx,cury
'' Print a zero-terminated string
  repeat strsize(stringptr)
    curx := ILIChar(byte[stringptr++],curx,cury)


PUB ILISetCursor(x,y)            
    curx := x
    cury := y


PUB ReadRamLong(WordStart,Bytevalue)                   ' easier calculating
    docmd("T",@result,Wordstart+Bytevalue>>1,2)
    return result

 
PUB ILILoadFont(stringptr) | filesize,i                ' read a .ifn file premade in angelcode bmfont and then the vb.net program
'    ' pass file name in stringptr
    i := fonttable                                      ' store at the constant 'fonttable' location, 2 bytes per address
    fat.openfile(stringptr,"R")                         ' open an image. Size is in the first long and has been rounded up to the next 256 byte block
    filesize := fat.filesize                            ' read the file size. This file has been rounded up to the next 256 byte block
    repeat (filesize >> 8)                              ' read in this number of 256 byte blocks
      fat.readdata(@rambuffer,256)                      ' get 256 bytes from sd card
      HubToRam(i,128)                                   ' 128 words = 256 bytes 
      i += 128                                          ' counter for ram location
    fat.closefile
    docmd("T",@rambuffer,FontTable,128)                 ' read first 128 words =256 bytes back
    FontHeight := byte[@rambuffer][251]                 ' font height
    'hex(fontheight,8)
    repeat i from 0 to 2
      sdbuffer[i] := byte[@rambuffer][35+i]
    docmd("E",@sdbuffer,@buffer2,1)                     ' convert to 2 byte .ili format
    BackFontColor := ( byte[@buffer2][1] << 8) | byte[@buffer2][0] ' correct order
    'hex(backfontcolor,4)    
    result := filesize

PUB DrawDesktop
    ' display picture
    Draw(0,0,screenwidth,screenheight)  ' dump out 76800 words to the display = a full screen
    RamToDisplay(Desktop,76800)                         ' picture is a global variable

PUB DrawPicture
    Draw(0,0,screenwidth,screenheight)  ' dump out 76800 words to the display = a full screen
    RamToDisplay(Picture,76800)                         ' picture is a global variable

PUB DrawSmallPicture(startx,starty,width,height,ramaddress)           ' draw a small picture
     Draw(startx,starty,startx+width-1,starty+height-1)
     RamToDisplay(ramaddress,width*height)      

PUB LoadRawPicture(stringptr,ramaddress)                ' take a standard RGB .raw file and convert to two byte RRRRRGGGGGGBBBBB format and fill memory
    fat.openfile(stringptr,"R")                         ' a .raw file 240 wide, 320 high
    repeat 600
      fat.readdata(@sdbuffer,384)                       ' 240x320=76800 bytes x3
      docmd("E",@sdbuffer,@rambuffer,128)               ' convert 3 bytes raw to 2 byte ili
      HubToRam(ramaddress,128)                          ' send to external ram
      ramaddress += 128                                 ' increment external ram counter
    fat.closefile                                       ' close file

PUB ReadMask | i,j                                            ' read 59*60 mask file into ram
    fat.openfile(string("mask.raw"),"R")
    i := Mask                                           ' location external ram of raw mask file
    repeat 60                                           ' 60 rows
      fat.readdata(@sdbuffer,59)                        ' get 59 bytes, do in pieces
      repeat j from 0 to 58
        word[@rambuffer][j] := sdbuffer[j]              ' convert a byte to a word
      docmd("S",@rambuffer,i,59)                        ' send to external ram  
      i +=59                                            ' increment counter 
    fat.closefile

PUB LoadSmallPicture(stringptr,width,height,ramaddress)  ' load a small .raw picture into ram (max size 8192 pixels)
    fat.openfile(stringptr,"R")
    repeat height
      fat.readdata(@sdbuffer,width*3)                   ' read the .raw file
      docmd("E",@sdbuffer,@rambuffer,width)           ' convert 3 bytes raw to 2 byte ili  
      HubToRam(ramaddress,width)                      ' send to ram
      ramaddress += width
    fat.closefile  

PUB ReadIcon(stringptr) | i
    fat.openfile(stringptr,"R")
    i := Icon                                           ' location in external ram of raw icon file
    repeat 60
      fat.readdata(@sdbuffer,177)                       ' get 59x3 bytes RGB values
      docmd("E",@sdbuffer,@rambuffer,59)                ' convert to RRRRRGGG_GGGBBBBB 59 words to rambuffer
      HubToRam(i,59)                                    ' send to external ram
      i +=59                                            ' add 59
    fat.closefile

PUB MergeBackgroundIcon(stringptr,x,y) | i,j,k,n                                 ' merge icon and background - if mask =0 then background, else icon (not classic alpha compositing but close enough)   
     ' i is background counter, j is icon counter, k = mask counter
    ReadIcon(stringptr)                                 ' read the icon off the sd card
    i := y*(screenwidth+1)+x                            ' location of picture pixel
    j := Icon                                           ' location of icon
    k := Mask                                           ' location of mask
    repeat 60
      docmd("T",@sdbuffer,i,59)                         ' get 59 pixels from background
      docmd("T",@rambuffer,j,59)                        ' get 59 icon pixels
      docmd("T",@buffer2,k,59)                          ' get words for the mask
      repeat n from 0 to 58                                         ' repeat for these pixels
        if word[@buffer2][n] >128
           word[@sdbuffer][n] :=word[@rambuffer][n]     ' if >128 then replace with icon pixel 
      docmd("S",@sdbuffer,i,59)                         ' send back modified line to the picture buffer    
      i += (screenwidth+1)                              ' next line in picture
      j +=59                                            ' icon next line
      k +=59                                            ' next line in mask (bytes) - buffer2
    DrawDesktop                              ' display modified picture

PUB HubToRam(RamAddress,Number)                         ' send pixels from hub to ram
   docmd("S",@rambuffer,RamAddress,Number)

PUB RamToHub(RamAddress,Number)                         ' send pixels from ram to hub
   docmd("T",@rambuffer,RamAddress,Number)

PUB RamToDisplay(RamAddress,Number)                     ' send pixels from ram to display
   docmd("U",0,RamAddress,Number)

Pub HubToDisplay(HubAddress,Number)                     ' send pixels from hub to display
   docmd("V",HubAddress,0,Number)   

PUB Clearscreen(color)                                  ' uses 2 byte ILI color (see below for rgb conversion)
    wordfill(@rambuffer,color,256)
    HubToRam(0,256)                                     ' send 256 pixels from hub to ram
    Draw(0,0,screenwidth,screenheight)                           ' set up the area of the screen to draw in   
    repeat 300
      RamToDisplay(0,256)

PUB RGBtoWord(R,G,B)                                     ' convert RGB to 2 byte ILI color 
    sdbuffer[0] := R
    sdbuffer[1] := G
    sdbuffer[2] := B                        
    docmd("E",@sdbuffer,@buffer2,1)
    result := ( byte[@buffer2][1] << 8) | byte[@buffer2][0] ' correct order 
    
PUB ClearRectangle(x1,y1,x2,y2,ColorWord) | i,size,remainder ' clears an area of the screen to the current font background color
    size := (x2-x1+1) * (y2 - y1 +1)                    ' number of pixels
    remainder := size & 255                             ' if not a whole number of 256 pixels
    repeat i from 0 to 255                              ' move the background font colour to the buffer
      rambuffer[i] := ColorWord
    Draw(x1,y1,x2,y2)                                   ' set up the area of the screen to draw in
    repeat size >> 8                                    ' 256 pixel blocks
      docmd("V",@rambuffer,0,256)                
    if remainder <> 0  
      docmd("V",@rambuffer,0,remainder)                 ' do the remainder
      
PUB SetOrientation(e) ' change orientation true = portrait. Changes global variable 'orientation' and also screen width and height
   orientation := e
   ChangeOrientation(orientation)
   if orientation
     screenwidth :=239
     screenheight :=319
   else
     screenwidth :=319
     screenheight :=239

PUB ChangeOrientation(n) ' pass true = portrait or false = landscape, changes global variable orientation in this object
' command $0001
'8.2.4. Driver Output Control (R01h)
'R/W RS D15 D14 D13 D12 D11 D10 D9 D8 D7 D6 D5 D4 D3 D2 D1 D0
'W 1 0 0 0 0 0 SM 0 SS 0 0 0 0 0 0 0 0
'SS: Select the shift direction of outputs from the source driver.
'When SS = 0, the shift direction of outputs is from S1 to S720
'When SS = 1, the shift direction of outputs is from S720 to S1.
'In addition to the shift direction, the settings for both SS and BGR bits are required to change the
'assignment of R, G, B dots to the source driver pins.
'To assign R, G, B dots to the source driver pins from S1 to S720, set SS = 0.
'To assign R, G, B dots to the source driver pins from S720 to S1, set SS = 1.

' command $0003
 'R/W RS D15 D14 D13 D12 D11 D10 D9  D8 D7  D6   D5   D4   D3 D2 D1 D0
   'W   1  TRI DFM 0   BGR 0   0   HWM 0  ORG 0    I/D1 I/D0 AM 0  0  0
   ' When AM = 0, the address is updated in horizontal writing direction.
   'When AM = 1, the address is updated in vertical writing direction.
   'I/D[1:0] Control the address counter (AC) to automatically increase or decrease by 1 when update one pixel
   ' display data.
    SelectMemGroup                                      ' 138 2 and appropriate pins
    orientation := n
    if orientation
      ' for origin top left and portrait mode
      ILIcmd($0001,%00000001_00000000)                               ' $0001,$0100 set SS and SM bit 0001 0100   portrait
      ILIcmd($0003,%00010000_00110000)                               ' $0003,$1030 set GRAM write direction and BGR=1. $0003 $1030  
    else
      ' for origin top right and landscape mode 
      ILIcmd($0001,%00000000_00000000)                               ' set SS and SM bit 0001 0000   landscape   
      ILIcmd($0003,%00010000_00111000)                  ' landscape  $1028 = original but 1038 is correct - not mirror image


PUB Change138(n)                                        ' just change the 138 do nothing else. pass n=0-7
   result := OUTA                                       ' get the current value of OUTA
   result &= %11110001_11111111_11111111_11111111         ' set P25-P27 low
   n := n<<25                                           ' shift the number
   result |= n                                          ' combine with n
   OUTA := result                                        ' send out
   DIRA |= %00001110_00000000_00000000_00000000         ' set the dira pins so P25-P27 are outputs


PUB DeselectGroup                                       ' 
   Change138(7)                                         ' deselect 138
   spi.stop                                             ' stop the spi cog so have control over these pins
                                                        ' stop any other cogs that might have control over the pins 
   DIRA &= %11111111_11100000_00000000_00000000         ' P0-P20 all are now inputs
   
PUB SelectSPIGroup                                      ' select group 0 for the SPI
   OUTA |= %00000000_00000000_00000000_00000000         ' P0 clock high
   DIRA &= %11111111_11100000_00000000_00000000         ' P0-P20 all inputs (SPI routine sets pin value and direction)
   Change138(0)                                         ' select group
   TouchStart                                           ' start the spi controller

'PUB SelectLoadGroup                                     ' handled in pasm now   

PUB SelectMemGroup                                      ' select group 2 for memory transfer
   DeselectGroup                                        ' turns off spi cog and deselects 138 outputs
   OUTA |= %00000000_00011111_11111111_11111111         ' pins P0-P20 high
   DIRA |= %00000000_00011111_11111111_11111111         ' P0-P20 as outputs
   Change138(2)



PUB Propfont_string(stringptr)                          'print at curx,cury
  repeat strsize(stringptr)
    Propfont_out(byte[stringptr++])
  crlf

PUB crlf
    curx := 0
    cury += 32                                          ' new line at end of string
    if cury >319                                        ' bottom of screen so new screen
      curx:=0
      cury:=0

PUB Propfont_out(ascii) | address,pixels
    Draw(curx,cury,curx+15,cury+31)                     ' location to start drawing 
    address := $8000 + (ascii >> 1) << 7                ' get rom address
    repeat 32                                           ' 32 rows per character, split in two parts
      pixels := long[address]                           ' get rom font data
      pixels := pixels >> (ascii & 1)                   ' shift for odd characters
      repeat 16 ' 16 columns
        if pixels & 1
          Pixel(%00000111_11100000)                     ' foreground color RRRRRGGG_GGGBBBBB
        else
          Pixel(%00000000_00000000)                     ' background color 
        pixels := pixels >> 2                           ' alternate pixels interleaved so shift 2
      address += 4 
    curx +=16
    if curx >239                                        ' new line
      crlf

PUB Start_ILI9325 ' pass orientation true = portrait, false = landscape
    ILI_Reset_High
    pause1ms(5)
    ILI_Reset_Low
    pause1ms(5)
    ILI_Reset_High
    ILI_CS_High
    ILI_RD_High
    ILI_WR_High
    pause1ms(5)
    ILI_CS_Low
          ' ************* Start Initial Sequence **********
    ILIcmd($00E5,$78F0)                                 ' set SRAM internal timing
    ILIcmd($0001,$0100)                                 ' set SS and SM bit 0001 0100   portrait
    ILIcmd($0002,$0700)                                 ' set 1 line inversion 
    ILIcmd($0003,$1030)                                 ' set GRAM write direction and BGR=1. $0003 $1030
    ILIcmd($0004,$0000)                                 ' Resize register 
    ILIcmd($0008,$0207)                                 ' set the back porch and front porch 
    ILIcmd($0009,$0000)                                 ' set non-display area refresh cycle ISC[3:0]  
    ILIcmd($000A,$0000)                                 ' FMARK function  
    ILIcmd($000C,$0000)                                 ' RGB interface setting   
    ILIcmd($000D,$0000)                                 ' Frame marker Position   
    ILIcmd($000F,$0000)                                 ' RGB interface polarity  
'        *************Power On sequence ****************//
    ILIcmd($0010,$0000)                                 ' SAP, BT[3:0], AP, DSTB, SLP, STB 
    ILIcmd($0011,$0007)                                 ' DC1[2:0], DC0[2:0], VC[2:0] 
    ILIcmd($0012,$0000)                                 ' VREG1OUT voltage   
    ILIcmd($0013,$0000)                                 ' VDV[4:0] for VCOM amplitude  
    ILIcmd($0007,$0001)   
    pause1ms(50)                                        ' Dis-charge capacitor power voltage
    ILIcmd($0010,$1090)                                 ' 1490//SAP, BT[3:0], AP, DSTB, SLP, STB   
    ILIcmd($0011,$0227)                                 ' DC1[2:0], DC0[2:0], VC[2:0] 
    pause1ms(50)                                        ' delay
    ILIcmd($0012,$001F)                                 ' 001C// Internal reference voltage= Vci;
    pause1ms(50)                                        ' delay   
    ILIcmd($0013,$1500)                                 ' $1000//1400   Set VDV[4:0] for VCOM amplitude  1A00  
    ILIcmd($0029,$0027)                                 ' $0012 //001a  Set VCM[5:0] for VCOMH  //$0025  0034  
    ILIcmd($002B,$000D)                                 ' Set Frame Rate   000C       
    pause1ms(50)                                        ' delay   
    ILIcmd($0020,$0000)                                 ' GRAM horizontal Address   
    ILIcmd($0021,$0000)                                 ' GRAM Vertical Address    
'         ----------- Adjust the Gamma Curve ----------//
    ILIcmd($0030,$0000) 
    ILIcmd($0031,$0707)  
    ILIcmd($0032,$0307)  
    ILIcmd($0035,$0200)  
    ILIcmd($0036,$0008)  
    ILIcmd($0037,$0004)  
    ILIcmd($0038,$0000)  
    ILIcmd($0039,$0707)  
    ILIcmd($003C,$0002)  
    ILIcmd($003D,$1D04)  
'        ------------------ Set GRAM area ---------------//
    ILIcmd($0050,$0000)                                 ' Horizontal GRAM Start Address
    ILIcmd($0051,$00EF)                                 ' Horizontal GRAM End Address       
    ILIcmd($0052,$0000)                                 ' Vertical GRAM Start Address    
    ILIcmd($0053,$013F)                                 ' Vertical GRAM Start Address       
    ILIcmd($0060,$A700)                                 ' Gate Scan Line     
    ILIcmd($0061,$0001)                                 ' NDL,VLE, REV    
    ILIcmd($006A,$0000)                                 ' set scrolling line      
'        -------------- Partial Display Control ---------/
    ILIcmd($0080,$0000)
    ILIcmd($0081,$0000)
    ILIcmd($0082,$0000)
    ILIcmd($0083,$0000)
    ILIcmd($0084,$0000)
    ILIcmd($0085,$0000)
'        //-------------- Panel Control -------------------//
    ILIcmd($0090,$0010)
    ILIcmd($0092,$0600)
    ILIcmd($0007,$0133)                                 ' 262K color and display ON
    ChangeOrientation(true)                             ' default to portrait

PUB Draw(x1, y1, x2, y2)                                ' sets the pixel to x1,y1 and then fills the next (x2-x1)*(y2-y1) pixels
    SelectMemGroup                                      ' set 138 to correct value
    ifnot orientation                                   ' landscape mode so swap x and y
      result :=x1                                       ' swap x1 and y1
      x1 := y1
      y1 := result
      result := x2                                      ' swap x2 and y2
      x2 :=y2
      y2 := result
    ILIcmd($0050,x1)
    ILIcmd($0052,y1)
    ILIcmd($0051,x2)
    ILIcmd($0053,y2)
    ILIcmd($0020,x1) 
    ILIcmd($0021,y1)
    Lcd_Write_Com($0022)


PUB Pixel(pixelcolor)               ' send out a pixel
    Lcd_Write_Data(pixelcolor)

PUB pause1ms(period)

'' Pause execution for period (in units of 1 ms).

  clkcycles := ((clkfreq / _1ms * period) - 4296) #> 381     ' Calculate 1 ms time unit
  waitcnt(clkcycles + cnt)                                   ' Wait for designated time

PUB hex(value, digits)   ' uses internal prop font and slow pixel drawing (for very basic debugging without an SD font)
  SelectMemGroup                               ' reselect 138 for display output for debugging  
'' Print a hexadecimal number
  propfont_out("O")
  propfont_out("x")
  value <<= (8 - digits) << 2
  repeat digits
    propfont_out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))
  crlf
                   

PRI ILIcmd(c,d)  ' instruction in one method
    Lcd_Write_Com(c) ' send out a word
    Lcd_Write_Data(d)

PRI Lcd_Write_Com(ILIlong)
    ILI_RS_low
    LCD_Writ_Bus(ILIlong)
    'docmd("W",ILIlong,0,0)  ' try moving to pasm, leaves spin and cog dira as 0 so need to reset both
    'SelectMemGroup           ' not working though. Not sure if should be ILIlong or @ILIlong. neither work
    

PRI Lcd_Write_Data(ILIlong)
    ILI_RS_High
    LCD_Writ_Bus(ILIlong)    

PRI LCD_Writ_Bus(ILILong)
    'ILILong &= %00000000_00000000_11111111_11111111    ' mask to a word. Not needed if care taken always to send a word value
    OUTA &= %11111111_11111111_00000000_00000000        ' set P0-P15 to zero ready to OR
    OUTA |= ILILong                                     ' merge with the word to output
    ILI_WR_Low                                          ' send out the data
    ILI_WR_High

PRI ILI_RS_Low
   OUTA &= %11111111_11111011_11111111_11111111         ' P18 low

PRI ILI_RS_High   
   OUTA |= %00000000_00000100_00000000_00000000         ' P18 high

PRI ILI_RD_Low                                          ' pulled high in hardware, not used  
PRI ILI_RD_High                                           

PRI ILI_WR_Low
   OUTA &= %11111111_11111101_11111111_11111111         ' p17 low

PRI ILI_WR_High
   OUTA |= %00000000_00000010_00000000_00000000         ' p17 high

PRI ILI_CS_Low                                          ' not used as setting 138 to %010 does this
PRI ILI_CS_High   

PRI ILI_RESET_Low                                       ' reset low
   Change138(3)                                         ' reset low

PRI ILI_RESET_High
   Change138(2)                                         ' 138 back to memselect  

DAT
sdbuffer        byte    $0[512]  ' 512 byte buffer for sd card interface
buffer2         byte    $0[512]  ' 512 general purpose hub buffer
rambuffer       word    $0[256]  ' 512 byte (256 word) buffer easier for using with words  

PUB start_ram : err_
' Initialise the Drac Ram driver. No actual changes to ram as the read/write routines handle this
  command := "I"
  cog := 1 + cognew(@tbp2_start, @command)
  if cog == 0
    err_ := $FF                 ' error = no cog
  else
    repeat while command        ' driver cog sets =0 when done
    err_ := errx                ' driver cog sets =0 if no error, else xx = error code

PUB stop_ram
   if cog
      cogstop(cog~ - 1)      

PUB DoCmd(command_, hub_address, ram_address, block_length) : err_
' Do the command: A-Z (I is reserved for Initialise)
  dira := 0                     ' tristate all pins with the spin dira
  hubaddrs := hub_address       ' hub address start
  ramaddrs := ram_address       ' ram address start
  blocklen := block_length      ' block length
  command  := command_          ' must be last !!
' Wait for command to complete and get status
  repeat while command          ' driver cog sets =0 when done
  err_ := errx                  ' driver cog sets =0 if no error, else xx = error code

CON
''Dracblade driver for talking to a ram chip via three latches
'' Modified code from Cluso's triblade
'' also includes commands to move blocks of data to the ILI9325 touchscreen display
' DoCmd(command_, hub_address, ram_address, block_length)
' I - initialise     
' S - Move data from hub to ram
' T - Move data from ram to hub
' U - Move data from ram to display

VAR

' communication params(5) between cog driver code - only "command" and "errx" are modified by the driver
   long  command, hubaddrs, ramaddrs, blocklen, errx, cog ' rendezvous between spin and assembly (can be used cog to cog)
'        command  = A to Z etc =0 when operation completed by cog
'        hubaddrs = hub address for data buffer
'        ramaddrs = ram address for data
'        blocklen = ram buffer length for data transfer
'        errx     = returns =0 (false=good), else <>0 (true & error code)
'        cog      = cog no of driver (set by spin start routine)
   

DAT
'' +-----------------------------------------------------------------------------------------------+
'' | Touchblade 161 Ram Driver (with grateful acknowlegements to Cluso and Average Joe)            |
'' +-----------------------------------------------------------------------------------------------+
                        org     0
tbp2_start    ' setup the pointers to the hub command interface (saves execution time later
                                      '  +-- These instructions are overwritten as variables after start
comptr                  mov     comptr, par     ' -|  hub pointer to command                
hubptr                  mov     hubptr, par     '  |  hub pointer to hub address            
ramptr                  add     hubptr, #4      '  |  hub pointer to ram address            
lenptr                  mov     ramptr, par     '  |  hub pointer to length                 
errptr                  add     ramptr, #8      '  |  hub pointer to error status           
cmd                     mov     lenptr, par     '  |  command  I/R/W/G/P/Q                  
hubaddr                 add     lenptr, #12     '  |  hub address                           
ramaddr                 mov     errptr, par     '  |  ram address                           
len                     add     errptr, #16     '  |  length                                
err                     nop                     ' -+  error status returned (=0=false=good) 


' Initialise hardware tristates everything and read/write set the pins
init                    mov     err, #0                  ' reset err=false=good
                        mov     dira,zero                ' tristate the pins with the cog dira

done                    wrlong  err, errptr             ' status  =0=false=good, else error x
                        wrlong  zero, comptr            ' command =0 (done)
' wait for a command (pause short time to reduce power)
pause
'                        mov     ctr, delay      wz      ' if =0 no pause
'              if_nz     add     ctr, cnt
'              if_nz     waitcnt ctr, #0                 ' wait for a short time (reduces power)
                        rdlong  cmd, comptr     wz      ' command ?
              if_z      jmp     #pause                  ' not yet
' decode command
                        cmp     cmd, #"S"       wz      ' hub to ram
              if_z      jmp     #pasmhubtoram           
                        cmp     cmd, #"T"       wz      ' ram to hub
              if_z      jmp     #pasmramtohub
                        cmp     cmd, #"U"       wz      ' ram to display
              if_z      jmp     #pasmramtodisplay
                        cmp     cmd, #"V"       wz      ' hub to display
              if_z      jmp     #pasmhubtodisplay           
                        cmp     cmd, #"E"       wz      ' convert 3 byte .raw format to 2 byte .ili format - hub to hub
              if_z      jmp     #rawtoiliformat
 '                       cmp     cmd, #"W"       wz      ' lcdwritecom in pasm, not working
 '             if_z      jmp     #pasmlcdwritecom
                        cmp     cmd, #"I"       wz      ' init
              if_z      jmp     #init     
                        mov     err, cmd                ' error = cmd (unknown command)
                        jmp     #done
                        
' ----------------- common routines -------------------------------------

get_values              rdlong  hubaddr, hubptr         ' get hub address
                        rdlong  ramaddr, ramptr         ' get ram address
                        rdlong  len, lenptr             ' get length
                        mov     err, #5                 ' err=5
get_values_ret          ret

                    ' ??come to this with possibly all pins tristated so need to make P16-P20 high before changing the 138 value 
set138                  shl     pasm_n,#25              ' pass n =0 to 7
                        or      dira,maskP0P20P25P27    ' make P25-P27 outputs as well as P0 to P20
                        andn    outa,mask138            ' make these three pins low
                        or      outa,pasm_n             ' set the 138 pins
set138_ret              ret


load161pasm                                             ' uses ramaddr
                        mov     pasm_n,#7
                        call    #set138                 ' deselect previous 138 value
                        or      dira,maskP0P20          ' %00000000_00011111_11111111_11111111         ' P0-P18 enabled for output plus P19,P20 
                        and     outa,maskP18low         ' %00001111_11111000_00000000_00000000         ' preserve previous values but set A0-18 low   
                        or      outa,ramaddr            ' output address to 161 chips
                        andn    outa,maskP19            ' set pin 19 low =  161 clock
                        mov     pasm_n,#1               ' 161 load low
                        call    #set138                 ' set it low
                        or      outa,maskP19            ' set pin 19 high = 161 clock
                        or      outa,maskP16P20         ' %00000000_00011111_00000000_00000000         ' set P16-P20 high prior to changing 138 
                        mov     pasm_n,#2               ' 161 load high and back to mem transfer
                        call    #set138                 ' send out
load161pasm_ret         ret

stop                   jmp     #stop                  ' for debugging



' ------------------ single letter commands  -------------------------------------
' command S
pasmhubtoram            call    #get_values             ' get hubaddr,ramaddr,len
                        call    #load161pasm                ' load the 161 counters with ramaddr
hubtoram_loop           and     outa,maskP16P31         '%11111111_11111111_00000000_00000000       ' clear for output                   
                        rdword  data_16,hubaddr         ' get the word from hub
                        and     data_16,maskP0P15       ' mask to a word only
                        or      outa,data_16            ' send out the byte to P0-P15
                        andn    outa,maskP20            ' set write low
                        add     hubaddr,#2              ' increment by 2 bytes = 1 word. Put this here for small delay while writes
                        or      outa,maskP20            ' write high
                        andn    outa,maskP19            ' clock 161 low
                        or      outa,maskP19            ' clock 161 high
                        djnz    len,#hubtoram_loop      ' loop this many times
                        jmp     #init                   ' tristate pins and listen for commands

' command T
pasmramtohub            call    #get_values             ' get hubaddr,ramaddr,len
                        call    #load161pasm            ' load the 161 counters with ramaddr
                        and     dira,maskP16P27         ' %00001111_11111111_00000000_00000000 set P0-P15 as inputs   
                        andn    outa,maskP16            ' memory /rd low
ramtohub_loop           mov     data_16,ina             ' get the data
                        wrword  data_16,hubaddr         ' move data to hub
                        andn    outa,maskP19            ' clock 161 low
                        or      outa,maskP19            ' clock 161 high
                        add     hubaddr,#2              ' increment the hub address 
                        djnz    len,#ramtohub_loop
                        or      outa,maskP16            ' memory /rd high  
                        or      dira,maskP0P15          ' %00000000_00000000_11111111_11111111 restore P0-P15as outputs
                        jmp     #init                   ' ' tristate pins and listen for commands

' command U
pasmramtodisplay        call    #get_values             ' get hubaddr,ramaddr,len
                        call    #load161pasm            ' load the 161 counters with ramaddr
                        or      outa,maskP18            ' ILI_RS high
                        andn    outa,maskP16            ' memory /rd low 
                        and     dira,maskP16P27         ' disable prop pins %00001111_11111111_00000000_00000000 set P0-P15 as inputs    
ramtodisplay_loop       andn    outa,maskP17            ' ILI write low
                        or      outa,maskP17            ' ILI write high
                        andn    outa,maskP19            ' clock 161 low
                        or      outa,maskP19            ' clock 161 high
                        djnz    len,#ramtodisplay_loop
                        or      outa,maskP16            ' memory /rd high  
                        or      dira,maskP0P15          ' %00000000_00000000_11111111_11111111 restore P0-P15as outputs
                        jmp     #init

' command V
pasmhubtodisplay        call    #get_values             ' get hubaddr,ramaddr,len
                        or      outa,maskP16P20         ' %00000000_00011111_00000000_00000000         ' set P16-P20 high prior to changing 138
                        mov     pasm_n,#7
                        call    #set138                 ' deselect previous 138 value
                        mov     pasm_n,#2               ' mem transfer
                        call    #set138                 ' send out
hubtodisplay_loop       and     outa,maskP16P31         '%11111111_11111111_00000000_00000000       ' clear for output                   
                        rdword  data_16,hubaddr         ' get the word from hub
                        and     data_16,maskP0P15       ' mask to a word only
                        or      outa,data_16            ' send out the byte to P0-P15
                        andn    outa,maskP17            ' ILI write low
                        or      outa,maskP17            ' ILI write high
                        add     hubaddr,#2              ' one word
                        djnz    len,#hubtodisplay_loop
                        jmp     #init

'command E
RawtoILIformat          ' takes a .raw 3 byte RRRRRRRR GGGGGGGG BBBBBBBB and converts to 2 byte RRRRRGGG GGGBBBBB
                        ' pass hubaddr, ramaddr and len
                        ' hubaddr is source location, len is number of pixels
                        ' ramaddr is destination in hub (messy naming) and length is 2/3 of blocklength
                        call    #get_values ' gets hubaddress, ramaddress and len (ignores ramaddress)
rawloop
                        rdbyte red,hubaddr
                        add hubaddr,#1
                        rdbyte green,hubaddr
                        add hubaddr,#1
                        rdbyte blue,hubaddr
                        add hubaddr,#1
                        call #rgbtoili
                        wrbyte ililow,ramaddr
                        add ramaddr,#1
                        wrbyte ilihigh,ramaddr
                        add ramaddr,#1
                        djnz    len,#rawloop            ' loop until done 
                        jmp     #init                   ' set pins to tristate

RGBtoILI                ' pass red,green, blue, returns ililow and ilihigh
                        shr     red,#3                  ' 000RRRRR 
                        shl     red,#3                  ' RRRRR000 
                        shr     green,#2                ' 00GGGGGG
                        mov     ilihigh,green           ' ilihigh = 00GGGGGG
                        shr     ilihigh,#3              ' ilihigh = 00000GGG
                        or      ilihigh,red             ' ilihigh = RRRRRGGG
                        and     green,#%00000111        ' 00000GGG
                        shl     green,#5                ' GGG00000
                        mov     ililow,green            ' ililow = GGG00000
                        shr     blue,#3                 ' blue = 000BBBBB
                        or      ililow,blue             ' ililow = GGGBBBBB
RGBtoILI_ret            ret

'pasmlcdwritecom         call    #get_values             ' use hubaddr as the data
'                        or      dira,maskP0P20          ' set these pins high (pass all pins tristated)
'                        or      outa,maskP0P20          '  set pins high
'                        mov     pasm_n,#2               '  mem transfer
'                        call    #set138                 ' set the 138
'                        andn    outa,maskP18            ' P18 ILIRS low
'                        and     outa,maskP16P31         ' set P0-P15 low
'                        or      outa,hubaddr            ' send out the data
'                        andn    outa,maskP17            ' ILI write low
'                        or      outa,maskP17            ' ILI write high
'                        jmp     #init                   ' set pins to tristate  

' variables
pasm_n                  long    0                                    ' general purpose value
data_16                 long    0                                    ' general purpose value
ililow                  long    0                                    ' low data byte latch5
ilihigh                 long    0                                    ' high data byte latch6
red                     long    0                                    ' red, green blue variables
green                   long    0
blue                    long    0           

' constants
Zero                    long    %00000000_00000000_00000000_00000000 ' used in several places
mask138                 long    %00001110_00000000_00000000_00000000 ' mask for the three 138 pins   
maskP0P20               long    %00000000_00011111_11111111_11111111 ' P0-P18 enabled for output plus P19,P20    
maskP18low              long    %00001111_11111000_00000000_00000000 ' P0-P18 low
maskP16                 long    %00000000_00000001_00000000_00000000 ' pin 16
maskP17                 long    %00000000_00000010_00000000_00000000 ' pin 17
maskP18                 long    %00000000_00000100_00000000_00000000 ' pin 18
maskP19                 long    %00000000_00001000_00000000_00000000 ' pin 19
maskP20                 long    %00000000_00010000_00000000_00000000 ' pin 20
maskP16P31              long    %11111111_11111111_00000000_00000000 ' pin 16 to pin 31
maskP0P20P25P27         long    %00001110_00011111_11111111_11111111  ' enable all pins as outputs except SD pins
maskP0P15               long    %00000000_00000000_11111111_11111111 ' for masking words
maskP16P20              long    %00000000_00011111_00000000_00000000
maskP16P27              long    %00001111_11111111_00000000_00000000
                        fit     496      